
# include "testing.h"


static void aux(void * pointer){
	if(pointer == NULL){
		ECHO(("____aux_before_THROW\n"));
		E4C_THROW(TamedException, "The external function will catch me.");
	}else{
		ECHO(("____aux_no_exception_was_thrown\n"));
	}
}

static int ext(void){

	volatile const		e4c_exception * thrown_exception;
	volatile E4C_BOOL	is_ready1;
	volatile E4C_BOOL	is_ready2;
	volatile int		status_code = -1;

	is_ready1 = e4c_context_is_ready();

	if(is_ready1){
		ECHO(("__ext_the_context_WAS_ready\n"));
	}else{
		ECHO(("__ext_the_context_WAS_NOT_ready\n"));
	}

	ECHO(("__ext_before_REUSING_CONTEXT\n"));

	{
		e4c_reusing_context(thrown_exception){

			ECHO(("__ext_before_TRY_block\n"));

			E4C_TRY{

				ECHO(("__ext_before_CALL_FUNCTION_aux\n"));

				aux(NULL);

				ECHO(("__ext_after_CALL_FUNCTION_aux\n"));

				status_code = 123;

			}E4C_CATCH(TamedException){

				ECHO(("__ext_inside_CATCH_block\n"));

				status_code = 0;

			}

			ECHO(("__ext_after_TRY_block\n"));

		}
	}

	ECHO(("__ext_after_REUSING_CONTEXT\n"));

	is_ready2 = e4c_context_is_ready();

	if(is_ready2){
		ECHO(("__ext_the_context_IS_ready\n"));
	}else{
		ECHO(("__ext_the_context_IS_NOT_ready\n"));
	}

	if(is_ready1 != is_ready2) return(112233);

	if(thrown_exception == NULL){

		ECHO(("__ext_there_was_no_error\n"));

		return(status_code);

	}else{

		ECHO(("__ext_there_was_an_error_%s\n", thrown_exception->name));

		return(status_code);
	}

}

DEFINE_TEST(
	h03,
	"A library (exception-unaware client) catches an exception",
	"This tests simulates a call to an external function (as in a library function). The client code is <em>exception-unaware</em>, but the external function uses the exception framework. So the external function opens a <code>e4c_reusing_context</code> and then a new exception context is created. The external function catches an exception thrown by another function and returns a status_code to its caller.",
	NULL,
	EXIT_SUCCESS,
	"result_was_0",
	NULL
){

	int result;

	ECHO(("before_CALL_FUNCTION_ext\n"));

	result = ext();

	ECHO(("after_CALL_FUNCTION_ext\n"));

	ECHO(("result_was_%d\n", result));

	return(EXIT_SUCCESS);
}
